/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * mazen-ng
 * Copyright (C) Jacob Zimmermann 2009 <jacob@jzimm.net>
 * 
 * mazen-ng is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * mazen-ng is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "advanced-selection.h"
#include "traversal-factory.h"
#include "dungeon-factory.h"
#include "castle-factory.h"
#include "diamond-factory.h"
#include "hex-factory.h"
#include "creta-factory.h"
#include <list>

using namespace Glib;
using namespace Gtk;

AdvancedSelection::AdvancedSelection(BaseObjectType* cobj, const RefPtr<Builder>& builder):
	Dialog (cobj), mazetypes () {
		// Factory list
		factories.push_back (new TraversalMazeFactory ());
		factories.push_back (new DungeonFactory ());
		factories.push_back (new CastleFactory ());
		factories.push_back (new DiamondFactory ());
		factories.push_back (new HexFactory ());
		factories.push_back (new CretaFactory ());
		
		// Factory selector
		HBox* mazetype_box;
		builder->get_widget ("mazetype_box", mazetype_box);

		// Populate the factory selector and choose a random factory
		std::vector<GridMazeFactory*>::iterator ifl;
		for (ifl = factories.begin (); ifl != factories.end(); ++ifl) {
			(*ifl)->trees_max_changed.connect (mem_fun (this, &AdvancedSelection::trees_max_changed));
			(*ifl)->trees_changed.connect (mem_fun (this, &AdvancedSelection::trees_changed));
			(*ifl)->floating_trees_max_changed.connect (mem_fun (this, &AdvancedSelection::floating_trees_max_changed));
			(*ifl)->floating_trees_changed.connect (mem_fun (this, &AdvancedSelection::floating_trees_changed));			
			mazetypes.append_text ((*ifl)->get_name ());
		}
		mazetypes.set_active (g_random_int_range (0, factories.size ()));

		// Put it in the window
		mazetype_box->add (mazetypes);
		mazetypes.show ();
		mazetypes.signal_changed ().connect (mem_fun (this, &AdvancedSelection::set_factory));

		// The various scales
		advanced_sizescale = RefPtr<HScale>::cast_static (builder->get_object ("advanced_sizescale"));
		sizescale = RefPtr<HScale>::cast_static (builder->get_object ("sizescale"));
		advanced_sizescale->set_range (10, 100);		
		sizescale->set_range (10, 100);

		ntrees = RefPtr<HScale>::cast_static (builder->get_object ("ntrees"));
		ntrees->signal_value_changed ().connect (mem_fun (this, &AdvancedSelection::ntrees_value_changed));

		floatingtrees = RefPtr<HScale>::cast_static (builder->get_object ("floatingtrees"));
		floatingtrees->signal_value_changed ().connect (mem_fun (this, &AdvancedSelection::ftrees_value_changed));

		sizescale->signal_value_changed ().connect (mem_fun (this, &AdvancedSelection::sizescale_value_changed));
		advanced_sizescale->signal_value_changed ().connect (mem_fun (this, &AdvancedSelection::advanced_sizescale_value_changed));

		set_factory ();
}

AdvancedSelection::~AdvancedSelection () {
	std::vector<GridMazeFactory*>::iterator ifl;
	for (ifl = factories.begin (); ifl != factories.end(); ++ifl)
		delete (*ifl);
}

void AdvancedSelection::sizescale_value_changed () {
	advanced_sizescale->set_value (sizescale->get_value ());
	get_factory ().set_size ((int) sizescale->get_value ());
}
void AdvancedSelection::advanced_sizescale_value_changed () {
	sizescale->set_value (advanced_sizescale->get_value ());
}

void AdvancedSelection::ntrees_value_changed () {
	get_factory ().set_trees ((int) ntrees->get_value ());
}

void AdvancedSelection::ftrees_value_changed () {
	get_factory ().set_floating_trees ((int) floatingtrees->get_value ());
}

void AdvancedSelection::set_factory () {
	GridMazeFactory& factory = get_factory ();
	sizescale->set_value ((double) factory.get_size ());
	double min_n_trees = (double) factory.get_min_trees ();
	double max_n_trees = (double) factory.get_max_trees ();
	double n_trees = (double) factory.get_trees ();
	double max_f_trees = (double) factory.get_max_floating_trees ();
	double f_trees = (double) factory.get_floating_trees ();

	ntrees->set_range (min_n_trees, max_n_trees);
	ntrees->set_value (n_trees);
	floatingtrees->set_range (0, max_f_trees);
	floatingtrees->set_value (f_trees);
}

void AdvancedSelection::trees_max_changed (int n) {
	ntrees->set_range ((double) get_factory ().get_min_trees (), (double) n);
}

void AdvancedSelection::trees_changed (int n) {
	ntrees->set_value ((double) n);
}

void AdvancedSelection::floating_trees_max_changed (int n) {
	floatingtrees->set_range (0, (double) n);
}

void AdvancedSelection::floating_trees_changed (int n) {
	floatingtrees->set_value ((double) n);
}