/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * mazen-ng
 * Copyright (C) Jacob Zimmermann 2009 <jacob@jzimm.net>
 * 
 * mazen-ng is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * mazen-ng is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma implementation

#include "artwork.h"
#include "util.h"
#include <librsvg/rsvg.h>
#include <librsvg/rsvg-cairo.h>
#include <string>
#include <algorithm>
#include <stdexcept>

using namespace std;
using namespace Cairo;

namespace Artwork {	
	class SvgImage {
		private:
			double width, height;
			RsvgHandle* svg;
			double usr_hmargin, usr_vmargin;
			double pict_w, pict_h, z_factor;
			
		public:	
			SvgImage (const string& svgfile);

			typedef unsigned long key;
			key get_key () const;
			
			void get_size (double& width, double& height) const;
			double p_width () const { return pict_w; }
			double p_height () const { return pict_h; }
			double p_scale () const { return z_factor; }
			double v_margin () const { return usr_vmargin; }
			double h_margin () const { return usr_hmargin; }
			
			void scale (RefPtr<Context> cr, const Rectangle& region);
			void place (RefPtr<Context> cr, const Rectangle& region, PositionRef pref, double& x, double& y) const;
			void render (RefPtr<Context> cr) const;
	};
	
	SvgImage::SvgImage (const string& svgfile) {
		GError* err = NULL;
		string fullpath = Resources::get_svg_path (svgfile);
		svg = rsvg_handle_new_from_file (fullpath.c_str (), &err);
		if (svg == NULL || err != NULL)
			throw runtime_error ("Could not load " + svgfile + "(" + fullpath + ")");

		RsvgDimensionData dim;
		rsvg_handle_get_dimensions (svg, &dim);
		pict_w = width = (double) dim.width;
		pict_h = height = (double) dim.height;
		usr_hmargin = usr_vmargin = 0;
		z_factor = 1;
	}

	inline SvgImage::key SvgImage::get_key () const {
		return (SvgImage::key) svg;
	}

	inline void SvgImage::get_size (double& width, double& height) const {
		width = this->width;
		height = this->height;
	}

	void SvgImage::scale (RefPtr<Context> cr, const Rectangle& region) {
		static const double dev_hmargin = 10;
		static const double dev_vmargin = 10;		

		usr_hmargin = dev_hmargin;
		usr_vmargin = dev_vmargin;
		cr->device_to_user_distance (usr_hmargin, usr_vmargin);

		double rw = region.width - (2 * usr_hmargin);
		double rh = region.height - (2 * usr_vmargin);

		z_factor = 1;
		double ph = height;
		
		if (width > rw) {
			z_factor = rw / width;
			ph *= z_factor;
		}
		if (ph > rh)
			z_factor *= (rh / ph);

		pict_w = width * z_factor;
		pict_h = height * z_factor;
	}

	void SvgImage::place (RefPtr<Context> cr, const Rectangle& region, PositionRef pref, double& x, double& y) const {
		switch (pref) {
			case LEFT:
				x -= (pict_w + usr_hmargin);
				y -= (pict_h / 2);
				break;
			case RIGHT:
				x += usr_hmargin;
				y -= (pict_h / 2);
				break;
			case TOP:
				x -= (pict_w / 2);
				y -= (pict_h + usr_vmargin);
				break;
			case BOTTOM:
				x = - (pict_w / 2);
				y += usr_vmargin;
				break;
			default:
				x -= (pict_w / 2);
				y -= (pict_h / 2);
		}
		// make sure we fit in the assigned rectangle
		x = CLAMP (x, usr_hmargin, region.width - usr_hmargin - pict_w);
		y = CLAMP (y, usr_vmargin, region.height - usr_vmargin - pict_h);
	}
	
	inline void SvgImage::render (RefPtr<Context> cr) const {
		rsvg_handle_render_cairo (svg, cr->cobj ());
	}

	struct _scene {
		SvgImage hero;
		SvgImage reward;

		_scene (const string& f_hero, const string& f_reward): hero (f_hero), reward (f_reward) {
		}
	};

	class SimpleRenderer: public Renderer {
		protected:
			virtual void render (SvgImage& svgi, Cairo::RefPtr<Cairo::Context> cr, const Cairo::Rectangle& region, double x, double y, PositionRef pref);

		public:
			void hero (art_handle h, Cairo::RefPtr<Cairo::Context> cr, const Cairo::Rectangle& region, double x, double y, PositionRef pref);
			void reward (art_handle h, Cairo::RefPtr<Cairo::Context> cr, const Cairo::Rectangle& region, double x, double y, PositionRef pref);
	};

	void SimpleRenderer::render (SvgImage& svgi, RefPtr<Context> cr, const Rectangle& region, double x, double y, PositionRef pref) {
		svgi.scale (cr, region);
		svgi.place (cr, region, pref, x, y);

		cr->save ();
		cr->translate (region.x + x, region.y + y);
		cr->scale (svgi.p_scale (), svgi.p_scale ());
		svgi.render (cr);
		cr->restore ();
	}

	void SimpleRenderer::hero (art_handle h, RefPtr<Context> cr, const Rectangle& region, double x, double y, PositionRef pref) {
		render (h->hero, cr, region, x, y, pref);
	}
	void SimpleRenderer::reward (art_handle h, RefPtr<Context> cr, const Rectangle& region, double x, double y, PositionRef pref) {
		render (h->reward, cr, region, x, y, pref);
	}

	class CachingRenderer: public SimpleRenderer {
		public:
			CachingRenderer ();
			
		private:
			typedef unsigned long cache_key;			
			static const cache_key dim_limit = 0xffff;

			cache_key key (double w, double h) const;

			typedef map<cache_key,RefPtr<ImageSurface> > imgcache;
			map<SvgImage::key,imgcache> cache;

			RefPtr<ImageSurface> lookup (const SvgImage& svgi, RefPtr<Context> cr, const Rectangle& region);
			RefPtr<ImageSurface> make_surface (const SvgImage& svgi, RefPtr<Context> cr, const Rectangle& region) const;

		protected:
			virtual void render (SvgImage& svgi, Cairo::RefPtr<Cairo::Context> cr, const Cairo::Rectangle& region, double x, double y, PositionRef pref);
	};

	CachingRenderer::CachingRenderer ():
		SimpleRenderer (),
		cache () {
	}

	CachingRenderer::cache_key CachingRenderer::key (double w, double h) const {
		cache_key kw = (cache_key) w, kh = (cache_key) h;
		kw &= dim_limit;
		kh &= dim_limit;

		return (kw << 16) | kh;
	}

	RefPtr<ImageSurface> CachingRenderer::make_surface (const SvgImage& svgi, RefPtr<Context> cr, const Rectangle& region) const {
		double w = svgi.p_width ();
		double h = svgi.p_height ();
		cr->user_to_device_distance (w, h);

		double iw, ih;
		svgi.get_size (iw, ih);

		double scale = w / iw;
		
		RefPtr<ImageSurface> sptr = ImageSurface::create (FORMAT_ARGB32, w, h);
		RefPtr<Context> icr = Context::create (sptr);
		icr->scale (scale, scale);
		svgi.render (icr);
		icr->stroke ();
		return sptr;
	}

	RefPtr<ImageSurface> CachingRenderer::lookup (const SvgImage& svgi, RefPtr<Context> cr, const Rectangle& region) {
		double rw = region.width;
		double rh = region.height;
		cr->user_to_device_distance (rw, rh);
		
		imgcache& icache = cache[svgi.get_key ()];		
		cache_key k = key (rw, rh);
		imgcache::const_iterator it = icache.find (k);		
		RefPtr<ImageSurface> surf;

		if (it == icache.end ()) { // not in the cache
			surf = make_surface (svgi, cr, region);
			if (region.width < (double) dim_limit && region.height < (double) dim_limit)
				icache[k] = surf;
		}
		else
			surf = it->second;
		return surf;
	}		

	void CachingRenderer::render (SvgImage& svgi, RefPtr<Context> cr, const Rectangle& region, double x, double y, PositionRef pref) {
		svgi.scale (cr, region);
		svgi.place (cr, region, pref, x, y);
		RefPtr<ImageSurface> surf = lookup (svgi, cr, region);

		double ox = region.x + x;
		double oy = region.y + y;
		cr->user_to_device (ox, oy);
		
		cr->save ();
		cr->set_identity_matrix ();
		cr->set_source (surf, ox, oy);
		cr->paint ();
		cr->restore ();
	}
	
	static SimpleRenderer _render;
	static CachingRenderer _render_cache;

	Renderer& render () { return _render; }
	Renderer& render_cache () { return _render_cache; }

	static vector<art_handle> pairs;
	static vector<art_handle>::iterator next_pair;

	static void load (const string& f_hero, const string& f_reward) {
		pairs.push_back (new _scene (f_hero, f_reward));
	}
	
	void init () {
		rsvg_init ();

		load ("walker.svg", "home.svg");
		load ("car.svg", "sailing.svg");
		load ("dog.svg", "bone.svg");
		load ("girl.svg", "cake.svg");
		load ("camper.svg", "fire.svg");
		load ("frog.svg", "mushroom.svg");
		load ("turbosnail.svg", "shamrock.svg");
		load ("rocket.svg", "planet.svg");
		load ("man_01.svg", "burger.svg");
		load ("frog.svg", "mushroom_benji_park_.svg");
		load ("mosquito.svg", "red_flower_01.svg");
		load ("present.svg", "christmas_tree_02.svg");
		load ("fire_engine_mimooh_01.svg", "firestation_mimooh_01.svg");
		load ("architect.svg", "caterpillar.svg");
		load ("signorinacappello.svg", "rosa2.svg");
		load ("mouse.svg", "cheese.svg");
		load ("pirateship.svg", "treasure-map.svg");
		load ("carlitos_Green_Worm.svg", "coredump_Peach.svg");

		next_pair = pairs.end (); // force scrambling
	}

	art_handle next () {
		if (next_pair == pairs.end ()) {
			random_shuffle (pairs.begin (), pairs.end ());
			next_pair = pairs.begin ();
		}
		return *(next_pair++);
	}
};
