/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * mazen-ng
 * Copyright (C) Jacob Zimmermann 2009 <jacob@jzimm.net>
 * 
 * mazen-ng is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * mazen-ng is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma implementation

#include "castle-maze.h"
#include <algorithm>
#include <stdexcept>

using namespace std;

// tower width or size
inline int get_tw (int t) {
	int tw = t / 3;
	if ((tw & 1) == 0)
		tw += 1;
	return tw;
}

inline int full_dim (int d) {
	int td = get_tw (d);
	return d + 2 * td;
}

CastleMaze::CastleMaze (int iwidth, int iheight, int n_trees, int n_floating_trees):
	DungeonMazeType (full_dim (iwidth), full_dim (iheight)) {
		if (iwidth < 10)
			throw runtime_error ("iwidth < 10 in CastleMaze");
		if (iheight < 10)
			throw runtime_error ("iheight < 10 in CastleMaze");
		
		cw = get_tw (iwidth);
		ch = get_tw (iheight);
		inner_width = iwidth;
		inner_height = iheight;		
		overlap_w = MAX (cw / 3, 2);
		overlap_h = MAX (ch / 3, 2);
		
		std::vector<xy> t;
		outer_wall_in_grid (t);
		towers (t);
		room_walls ();
		room_walls_in_grid (t);
		
		entrance.set (cw - overlap_w, height () / 2 - 1);
		clear_vert (entrance);
		set_no_no (exit);

		outer_wall ();		
		walls.push_back (outer);
		walls.push_back (roomh);
		walls.push_back (roomv);

		mark_no_no ();
		
		// create trees		
		random_shuffle (t.begin (), t.end ());
		for (int k = 0; k < n_trees; k++)
			tree_positions.push_back (t[k]);

		// create floating trees
		int x, y;
		for (int k = 0; k < n_floating_trees; k++) {
			x = g_random_int_range (ch - overlap_h + 1, inner_width - 2);
			y = g_random_int_range (cw - overlap_w + 1, inner_height - 2);
			if (is_clear (x, y))
				tree_positions.push_back (xy (x, y));
		}

		generate();
	}

void CastleMaze::mark_no_no () {
	for (int y = 0; y < ch - overlap_h; y++) {
		for (int x = cw; x < cw + inner_width - 1; x++) {
			set_no_no (x, y);
			set_no_no (x, height () - 1 - y);
		}
	}
	for (int y = ch; y < ch + inner_height - 1; y++) {
		for (int x = 0; x < cw - overlap_w; x++) {
			set_no_no (x, y);
			set_no_no (width () - 1 - x, y);
		}
	}

	// we don't want trees in the room either
	for (int x = cx; x < cx + cw; x++)
		for (int y = cy; y <= cy + ch; y++)
			set_no_no (x, y);
}

void CastleMaze::outer_wall_in_grid (vector<xy>& t) {
	xy p1, p2;
	for (int x = 0; x < inner_width - 1; x++) {
		p1.set (cw + x, ch - overlap_h);
		p2.set (cw + x, height () - ch - 1 + overlap_h);
		set_horiz (p1);
		set_horiz (p2);
		t.push_back (p1);
		t.push_back (p2);
	}	
	for (int y = 0; y < inner_height - 1; y++) {
		p1.set (cw - overlap_w, ch + y);
		p2.set (width () - cw - 1 + overlap_w, ch + y);
		set_vert (p1);
		set_vert (p2);
		if (y > 0 && y < inner_height - 2) {
			t.push_back (p1);  
			t.push_back (p2);
		}
	}	
}

void CastleMaze::towers (vector<xy>& t) {
	xy p[4];
	for (int x = 0; x < cw; x++) {
		p[0].set (x, 0);
		p[1].set (x, height () - 1);
		p[2].set (width () - 2 - x, 0);
		p[3].set (width () - 2 - x, height () - 1);
		for (int k = 0; k < 4; k++) {
			set_horiz (p[k]);
			if (x > 0 && x < cw - 1)
				t.push_back (p[k]);
		}
	}
	for (int y = 0; y < ch; y++) {
		p[0].set (0, y);
		p[1].set (width () - 1, y);
		p[2].set (0, height () - 2 - y);
		p[3].set (width () - 1, height () - 2 - y);
		for (int k = 0; k < 4; k++) {
			set_vert (p[k]);
			if (y > 0 && y < ch - 1)
				t.push_back (p[k]);
		}
	}
	for (int x = 0; x < cw - overlap_w; x++) {
		p[0].set (x, ch);
		p[1].set (width () - 2 - x, ch);
		p[2].set (x, height () - 1 - ch);
		p[3].set (width () - 2 - x, height () - 1 - ch);
		for (int k = 0; k < 4; k++) {
			set_horiz (p[k]);
			if (x > 0 && x < cw - overlap_w - 1)
				t.push_back (p[k]);
		}
	}
	for (int y = 0; y < ch - overlap_h; y++) {
		p[0].set (cw, y);
		p[1].set (cw, height () - 2 - y);
		p[2].set (width () - 1 - cw, y);
		p[3].set (width () - 1 - cw, height () - 2 - y);
		for (int k = 0; k < 4; k++) {
			set_vert (p[k]);
			if (y > 0 && y < ch - overlap_h - 1)
				t.push_back (p[k]);
		}		
	}
}

void CastleMaze::room_walls () {
	xy wcorner;
	int r_tower = g_random_int_range (0, 4);
	switch (r_tower) {
		case 0:		// north west
			exit.set (cw - overlap_w, ch);
			cx = cy = 0;
			wcorner = xy (cw, ch);
			roomh.push_back (exit.right ());
			roomh.push_back (wcorner);
			roomv.push_back (xy (cw, ch - overlap_h));
			roomv.push_back (wcorner);
			exit_dir = GridMaze::DOWN;
			break;
		case 1:		// north east
			exit.set (width () - cw, ch);
			cx = width () - cw - 1;
			cy = 0;
			wcorner = xy (cx, ch);
			roomh.push_back (wcorner);
			roomh.push_back (exit);
			roomv.push_back (xy (cx, ch - overlap_h));
			roomv.push_back (wcorner);
			exit_dir = GridMaze::DOWN;		
			break;
		case 2:		// south west
			exit.set (cw - overlap_w, height () - 1 - ch);
			cx = 0;
			cy = height () - ch - 1;
			wcorner = xy (cw, height () - 1 - ch);
			roomh.push_back (exit.right ());
			roomh.push_back (wcorner);
			roomv.push_back (wcorner);
			roomv.push_back (xy (cw, height () - 1 - ch + overlap_h));
			exit_dir = GridMaze::UP;
			break;
		case 3:		// south east
			exit.set (width () - cw, height () - 1 - ch);
			cx = width () - cw - 1;
			cy = height () - ch - 1;
			wcorner = xy (cx, height () - 1 - ch);
			roomh.push_back (wcorner);
			roomh.push_back (exit);
			roomv.push_back (wcorner);
			roomv.push_back (xy (cx, height () - 1 - ch + overlap_h));
			exit_dir = GridMaze::UP;
	}
}

void CastleMaze::room_walls_in_grid (vector<xy>& t) {
	Wall::const_iterator p = roomh.begin ();
	int wy = p->y, x1 = p->x;
	int x2 = (++p)->x;
	for (int x = x1; x < x2; x++) {
		set_horiz (x, wy);
		if (x > x1 && x < x2 - 1)
			t.push_back (xy (x, wy));
	}

	p = roomv.begin ();
	int wx = p->x, y1 = p->y;
	int y2 = (++p)->y;
	for (int y = y1; y < y2; y++) {
		set_vert (wx, y);
		if (y > y1 && y < y2 - 1)
			t.push_back (xy (wx, y));
	}
}

void CastleMaze::outer_wall () {
	outer.push_back (entrance);
	outer.push_back (xy (cw - overlap_w, ch));
	outer.push_back (xy (0, ch));
	outer.push_back (xy (0, 0));
	outer.push_back (xy (cw, 0));
	outer.push_back (xy (cw, ch - overlap_h));
	outer.push_back (xy (width () - cw - 1, ch - overlap_h));
	outer.push_back (xy (width () - cw - 1, 0));
	outer.push_back (xy (width () - 1, 0));
	outer.push_back (xy (width () - 1, ch));
	outer.push_back (xy (width () - cw + overlap_w - 1, ch));
	outer.push_back (xy (width () - cw + overlap_w - 1, height () - ch - 1));
	outer.push_back (xy (width () - 1, height () - ch - 1));
	outer.push_back (xy (width () - 1, height () - 1));
	outer.push_back (xy (width () - cw - 1, height () - 1));
	outer.push_back (xy (width () - cw - 1, height () - ch + overlap_h -1));
	outer.push_back (xy (cw, height () - ch + overlap_h -1));
	outer.push_back (xy (cw, height () - 1));
	outer.push_back (xy (0, height () - 1));
	outer.push_back (xy (0, height () - ch - 1));
	outer.push_back (xy (cw - overlap_w, height () - ch - 1));
	outer.push_back (entrance.up ());
}

xy CastleMaze::get_entrance () const { return entrance; }
xy CastleMaze::get_exit () const { return exit; }
