/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * mazen-ng
 * Copyright (C) Jacob Zimmermann 2009 <jacob@jzimm.net>
 * 
 * mazen-ng is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * mazen-ng is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma implementation

#include "mainwindow.h"
#include "grid-maze.h"
#include "grid-factory.h"
#include "grid-view.h"
#include "print.h"
#include "advanced-selection.h"
#include "util.h"
#include "config.h"
#include "export-dialog.h"

using namespace Glib;
using namespace Gtk;
using namespace sigc;
using namespace Cairo;

static const int history_capacity = 4096;

MainWindow::MainWindow (BaseObjectType* cobject, const Glib::RefPtr<Builder>& builder):
	Window (cobject) {	
		Glib::RefPtr<Gdk::Pixbuf> logo = Gdk::Pixbuf::create_from_file (Resources::get_logo_path ());
		set_icon (logo);
		
		Glib::RefPtr<Action> quit = Glib::RefPtr<Action>::cast_static (builder->get_object ("quit"));
		quit->signal_activate ().connect (mem_fun (this, &Window::hide));

		Glib::RefPtr<Action> next = Glib::RefPtr<Action>::cast_static (builder->get_object ("next"));
		next->signal_activate ().connect (mem_fun (this, &MainWindow::forward));

		previous = Glib::RefPtr<Action>::cast_static (builder->get_object ("previous"));
		previous->signal_activate ().connect (mem_fun (this, &MainWindow::back));
		previous->set_sensitive (false);
		
		Glib::RefPtr<Action> print_a = Glib::RefPtr<Action>::cast_static (builder->get_object ("print"));
		print_a->signal_activate ().connect (mem_fun (this, &MainWindow::print));		
		
		aboutmazen = Glib::RefPtr<AboutDialog>::cast_static (builder->get_object ("aboutmazen"));
		aboutmazen->set_version (PACKAGE_VERSION);
		aboutmazen->set_logo (logo);
		MenuItem* about;
		builder->get_widget ("about", about);
		about->signal_activate ().connect (mem_fun (this, &MainWindow::on_about));

		DrawingArea* _drawingarea;
		builder->get_widget ("drawingarea", _drawingarea);
		drawingarea = Glib::RefPtr<DrawingArea> (_drawingarea);
		drawingarea->signal_expose_event ().connect (mem_fun (this, &MainWindow::on_maze_redraw));
		drawingarea->signal_realize ().connect (mem_fun (this, &MainWindow::create));

		showsolution = Glib::RefPtr<ToggleAction>::cast_static (builder->get_object ("showsolution"));
		showsolution->signal_toggled ().connect (sigc::retype_return<void> (mem_fun (_drawingarea, &DrawingArea::queue_draw)));
		
		showadvanced = Glib::RefPtr<ToggleAction>::cast_static (builder->get_object ("showadvanced"));
		AdvancedSelection* _advanced;
		builder->get_widget_derived ("advanced", _advanced);
		advanced = Glib::RefPtr<AdvancedSelection> (_advanced);
		showadvanced->signal_activate ().connect (mem_fun (this, &MainWindow::create));

		MenuItem* save;
		builder->get_widget ("save", save);
		save->signal_activate ().connect (mem_fun (this, &MainWindow::save));

		ExportDialog* _edialog;
		builder->get_widget_derived ("exportdialog", _edialog);
		exportdialog = Glib::RefPtr<ExportDialog> (_edialog);

		maze_history.push_back (advanced->get_factory ().new_maze ());
		showsolution->set_active (false);
		previous->set_sensitive (false);
		current_view = maze_history.begin ();
}

void MainWindow::on_about () const {
	aboutmazen->show_all ();
	aboutmazen->run ();
	aboutmazen->hide ();
}

bool MainWindow::on_maze_redraw (GdkEventExpose* event) {
	Glib::RefPtr<Gdk::Window> da_window = drawingarea->get_window ();
	Cairo::RefPtr<Context> cr = da_window->create_cairo_context ();
	
	int w, h;
	da_window->get_size (w, h);
	if (w != view_width || h != view_height || (*current_view)->has_changed ()) {
		view_width = w;
		view_height = h;
		(*current_view)->set_size (w, h);
	}

	if ((*current_view)->has_changed ()) {
		Cairo::RefPtr<Surface> maze_surface = Surface::create (cr->get_target (), cr->get_target ()->get_content (), view_width, view_height);
		Cairo::RefPtr<Context> mcr = Context::create (maze_surface);
		// white background
		mcr->set_source_rgb (1, 1, 1);
		mcr->paint ();		
		(*current_view)->render (mcr, Artwork::render_cache ());
		cr->set_source (maze_surface, 0, 0);
		maze_pattern = cr->get_source ();

		Cairo::RefPtr<Surface> sol_surface = Surface::create (cr->get_target (), cr->get_target ()->get_content (), view_width, view_height);
		Cairo::RefPtr<Context> scr = Context::create (sol_surface);
		scr->set_source (maze_pattern);
		scr->paint ();
		(*current_view)->render_solution (scr, Artwork::render_cache ());
		cr->set_source (sol_surface, 0, 0);
		sol_pattern = cr->get_source ();
	}

	cr->rectangle (event->area.x, event->area.y, event->area.width, event->area.height);
	cr->clip ();
	if (showsolution->get_active ())
		cr->set_source (sol_pattern);
	else
		cr->set_source (maze_pattern);
	cr->paint ();

	return false;
}

void MainWindow::forward () {
	current_view ++;
	
	if (current_view == maze_history.end ()) {
		maze_history.push_back (advanced->get_factory ().new_maze ());
		showsolution->set_active (false);
	}
	else
		(*current_view)->force_redraw ();

	previous->set_sensitive (true);
	drawingarea->queue_draw ();
}

void MainWindow::back () {
	if (-- current_view == maze_history.begin ())
		previous->set_sensitive (false);

	(*current_view)->force_redraw ();
	drawingarea->queue_draw ();
}

void MainWindow::print () const {
	Print printjob (*current_view);
	printjob.run ();
}

void MainWindow::create () {
	if (showadvanced->get_active ()) {
		int res;
		do {
			res = advanced->run ();
		} while (res == 1);
		advanced->hide ();
		showadvanced->set_active (false);
	}
}

void MainWindow::save () const {
	exportdialog->save (*current_view, view_width, view_height);
}
